library(cluster)

mk.cluster.predictor = function(train, k,
        dists.f = function(train) 1 - cor(train, method = "kendall"),
#        dists = as.1 - cor(train, method = "pearson")
#       dists = sapply(colnames(train), function(v1) sapply(colnames(train), function(v2)
#           mean(abs(train[,v1] - train[,v2])))))
        train.model.f = function(x, y) lm(y ~ x),
        predict.from.model.f = function(model, x)
            num(predict(model, newdata = data.frame(x))))
   {if (any(is.na(train)))
       stop("'train' has a missing value")
#    if (any(mapcols(train, var) == 0))
#       stop("'train' has a 0-variance variable")
    dists = as.dist(dists.f(train))
#        as.dist(1 - cor(train, method = "kendall"))
    clusters = cutree(hclust(dists, method = "complete"), k = k)
    # Save means and SDs before standardizing 'train', so we can
    # apply the same transformations to new inputs in testing.
    means = colMeans(train)
    sds = mapcols(train, function(v) if (sdn(v) == 0) 1 else sdn(v))
    train.std = sapply(1 : ncol(train), function(i)
        (train[,i] - means[i])/sds[i])
    colnames(train.std) = colnames(train)
    models = lapply(colnames(train), function(v)
       {y = train[,v]
        others = setdiff(names(which(clusters == num(clusters[v]))), v)
        if (length(others))
           {x = (if (length(others) > 1) maprows(train.std[,others], median)
                else train.std[,others])
            train.model.f(x, y)}
        else
         # This is a singleton cluster, so predict this variable
         # with a constant model.
            median(y)})
    names(models) = colnames(train)
    punl(clusters, means, sds, models, predict.from.model.f)}

test.cluster.predictor = function(cp, test)
# Tests the given cluster predictor, using leave-one-out
# cross-validation to produce a prediction for every item in the
# test set. Returns a matrix of predictions.
   {test = `colnames<-`(
        sapply(1 : ncol(test), function(i)
            (test[,i] - cp$means[i])/cp$sds[i]),
        colnames(test))
    sapply(colnames(test), function(item)
        sapply(1 : nrow(test), function(s)
           {if (is.atomic(cp$models[[item]]))
              # It's a constant model.
                cp$models[[item]]
            else
              # It's a real model.
               {others = setdiff(
                    names(which(cp$clusters == num(cp$clusters[item]))),
                    item)
                x = median(test[s,][others])
                cp$predict.from.model.f(cp$models[[item]], x)}}))}

rmsd = function(x, y) sqrt(mean((x - y)^2))

cv.cluster.prediction = function(dataset, k, nfold = 10, ...)
# Tests the accuracy of cluster prediction on the given dataset with
# an outer round of cross-validation.
   {folds = sample(rep_len(1 : nfold, nrow(dataset)))
    pred = lapply(1 : nfold, function(fold)
       {cp = mk.cluster.predictor(dataset[which(folds != fold),], k = k, ...)
        test.cluster.predictor(cp, dataset[which(folds == fold),])})
    perm = order(c(recursive = T,
        lapply(1 : nfold, function(fold) which(folds == fold))))
    do.call(rbind, pred)[perm,]}

# go = function()
#    {pw = ss(prefwide, sb[row.names(prefwide), "gender"] == Male)
#     pw = pw[sapply(pref.items, function(pi)
#         all(!is.na(pw[pi])) && var(pw[pi]) > 0)]

remove.na.prefwide.cols = function(pw)
    pw[sapply(pref.items, function(pi) all(!is.na(pw[pi])))]

cv.cluster.prediction.prefs = function(prefwide, k)
    cv.cluster.prediction(prefwide, k = k,
        dists.f = function(train)
          # Deal with 0-variance variables by giving them
          # Kendall distance 0 to all other variables.
           {dm = 1 - cor(train, method = "kendall")
            dm[1:length(dm)] = ifelse(is.na(dm), 0, dm)},
        predict.from.model.f = function(model, x) pmax(1, pmin(7,
          # Clip predictions to [1, 7].
            num(predict(model, newdata = data.frame(x))))))

